"""CBOE Futures Curve Fetcher."""

# IMPORT STANDARD
from typing import Any, Dict, List, Optional

from openbb_cboe.utils.helpers import get_settlement_prices
from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.futures_curve import (
    FuturesCurveData,
    FuturesCurveQueryParams,
)
from pydantic import Field


class CboeFuturesCurveQueryParams(FuturesCurveQueryParams):
    """CBOE Futures Curve Query.

    Source: https://www.cboe.com/
    """


class CboeFuturesCurveData(FuturesCurveData):
    """CBOE Futures Curve Data."""

    symbol: str = Field(description="The trading symbol for the tenor of future.")


class CboeFuturesCurveFetcher(
    Fetcher[
        CboeFuturesCurveQueryParams,
        List[CboeFuturesCurveData],
    ]
):
    @staticmethod
    def transform_query(params: Dict[str, Any]) -> CboeFuturesCurveQueryParams:
        return CboeFuturesCurveQueryParams(**params)

    @staticmethod
    def extract_data(
        query: CboeFuturesCurveQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the CBOE endpoint."""

        query.symbol = query.symbol.upper()
        FUTURES = get_settlement_prices(settlement_date=query.date)
        if len(FUTURES) == 0:
            return []

        if query.symbol not in FUTURES["product"].unique().tolist():
            raise RuntimeError(
                "The symbol, "
                f"{query.symbol}"
                ", is not valid.  Chose from: "
                f"{FUTURES['product'].unique().tolist()}"
            )

        data = get_settlement_prices(settlement_date=query.date)
        data = data[data["product"] == query.symbol]

        return data[["expiration", "symbol", "price"]].to_dict("records")

    @staticmethod
    def transform_data(
        query: CboeFuturesCurveQueryParams,
        data: List[Dict],
        **kwargs: Any,
    ) -> List[CboeFuturesCurveData]:
        return [CboeFuturesCurveData.model_validate(d) for d in data]



"""CBOE European Index Constituents fetcher."""

from datetime import datetime
from typing import Any, Dict, List, Optional

import pandas as pd
from openbb_cboe.utils.helpers import EUR_INDEX_CONSTITUENTS_COLUMNS, Europe
from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.european_index_constituents import (
    EuropeanIndexConstituentsData,
    EuropeanIndexConstituentsQueryParams,
)
from openbb_provider.utils.helpers import make_request
from pydantic import Field, field_validator


class CboeEuropeanIndexConstituentsQueryParams(EuropeanIndexConstituentsQueryParams):
    """CBOE Stock end of day query.

    Source: https://www.cboe.com/

    Gets the current price data for all constituents of the CBOE European Index.
    """


class CboeEuropeanIndexConstituentsData(EuropeanIndexConstituentsData):
    """CBOE European Index Constituents Data.

    Current trading day price data for all constituents of the CBOE European Index.
    """

    prev_close: Optional[float] = Field(
        default=None, description="Previous closing  price."
    )
    change: Optional[float] = Field(default=None, description="Change in price.")
    change_percent: Optional[float] = Field(
        default=None, description="Change in price as a percentage."
    )
    tick: Optional[str] = Field(
        default=None, description="Whether the last sale was an up or down tick."
    )
    last_trade_timestamp: Optional[datetime] = Field(
        default=None, description="Last trade timestamp for the symbol."
    )
    exchange_id: Optional[int] = Field(
        default=None, description="The Exchange ID number."
    )
    seqno: Optional[int] = Field(
        default=None, description="Sequence number of the last trade on the tape."
    )
    asset_type: Optional[str] = Field(
        default=None, description="Type of asset.", alias="type"
    )

    @field_validator("last_trade_timestamp", mode="before", check_fields=False)
    def date_validate(cls, v):  # pylint: disable=E0213
        """Return the datetime object from the date string"""

        return datetime.strftime(v, "%Y-%m-%d %H:%M:%S")


class CboeEuropeanIndexConstituentsFetcher(
    Fetcher[
        CboeEuropeanIndexConstituentsQueryParams,
        List[CboeEuropeanIndexConstituentsData],
    ]
):
    """Transform the query, extract and transform the data from the CBOE endpoints"""

    @staticmethod
    def transform_query(
        params: Dict[str, Any]
    ) -> CboeEuropeanIndexConstituentsQueryParams:
        """Transform the query."""

        return CboeEuropeanIndexConstituentsQueryParams(**params)

    @staticmethod
    def extract_data(
        query: CboeEuropeanIndexConstituentsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the CBOE endpoint"""

        SYMBOLS = pd.DataFrame(Europe.list_indices())["symbol"].to_list()
        query.symbol = query.symbol.upper()

        if query.symbol not in SYMBOLS:
            raise RuntimeError(
                f"The symbol, {query.symbol},"
                "was not found in the CBOE European Index directory. "
                "Use `available_indices(europe=True)` to see the full list of indices."
            )
        url = f"https://cdn.cboe.com/api/global/european_indices/constituent_quotes/{query.symbol}.json"

        r = make_request(url)
        if r.status_code != 200:
            raise RuntimeError(r.status_code)

        r_json = r.json()

        data = (
            pd.DataFrame.from_records(r_json["data"])[
                list(EUR_INDEX_CONSTITUENTS_COLUMNS.keys())
            ]
            .rename(columns=EUR_INDEX_CONSTITUENTS_COLUMNS)
            .round(2)
        )
        data["last_trade_timestamp"] = pd.to_datetime(data["last_trade_timestamp"])

        return data.to_dict("records")

    @staticmethod
    def transform_data(
        query: CboeEuropeanIndexConstituentsQueryParams, data: List[Dict], **kwargs: Any
    ) -> List[CboeEuropeanIndexConstituentsData]:
        """Transform the data to the standard format"""

        return [CboeEuropeanIndexConstituentsData.model_validate(d) for d in data]



"""CBOE Helpers Module."""

from datetime import date, timedelta
from io import BytesIO, StringIO
from typing import Any, Dict, List, Optional

import pandas as pd
import requests
import requests_cache
from openbb_core.app.utils import get_user_cache_directory
from openbb_provider.utils.helpers import to_snake_case

cache_dir = get_user_cache_directory()
cboe_session = requests_cache.CachedSession(
    f"{cache_dir}/http/cboe_directories", expire_after=timedelta(days=7)
)

TICKER_EXCEPTIONS = ["NDX", "RUT"]

US_INDEX_COLUMNS = {
    "symbol": "symbol",
    "name": "name",
    "current_price": "price",
    "open": "open",
    "high": "high",
    "low": "low",
    "close": "close",
    "prev_day_close": "prev_close",
    "price_change": "change",
    "price_change_percent": "change_percent",
    "last_trade_time": "last_trade_timestamp",
}

EUR_INDEX_COLUMNS = {
    "name": "name",
    "isin": "isin",
    "currency": "currency",
    "current_price": "price",
    "open": "open",
    "high": "high",
    "low": "low",
    "close": "close",
    "prev_day_close": "prev_close",
    "price_change": "change",
    "price_change_percent": "change_percent",
    "last_trade_time": "last_trade_timestamp",
}

EUR_INDEX_CONSTITUENTS_COLUMNS = {
    "symbol": "symbol",
    "current_price": "price",
    "open": "open",
    "high": "high",
    "low": "low",
    "close": "close",
    "volume": "volume",
    "seqno": "seqno",
    "prev_day_close": "prev_close",
    "price_change": "change",
    "price_change_percent": "change_percent",
    "last_trade_time": "last_trade_timestamp",
    "exchange_id": "exchange_id",
    "tick": "tick",
    "security_type": "type",
}


def get_cboe_directory(**kwargs) -> pd.DataFrame:
    """Get the US Listings Directory for the CBOE.

    Returns
    -------
    pd.DataFrame: CBOE_DIRECTORY
        DataFrame of the CBOE listings directory
    """

    r = cboe_session.get(
        "https://www.cboe.com/us/options/symboldir/equity_index_options/?download=csv",
        timeout=10,
    )

    if r.status_code != 200:
        raise RuntimeError(r.status_code)

    r_json = StringIO(r.text)

    CBOE_DIRECTORY = pd.read_csv(r_json)

    CBOE_DIRECTORY = CBOE_DIRECTORY.rename(
        columns={
            " Stock Symbol": "symbol",
            " DPM Name": "dpm_name",
            " Post/Station": "post_station",
            "Company Name": "name",
        }
    ).set_index("symbol")

    return CBOE_DIRECTORY.astype(str)


def get_cboe_index_directory(**kwargs) -> pd.DataFrame:
    """Get the US Listings Directory for the CBOE

    Returns
    -------
    pd.DataFrame: CBOE_INDEXES
    """

    r = cboe_session.get(
        "https://cdn.cboe.com/api/global/us_indices/definitions/all_indices.json",
        timeout=10,
    )

    if r.status_code != 200:
        raise requests.HTTPError(r.status_code)

    CBOE_INDEXES = pd.DataFrame(r.json())

    CBOE_INDEXES = CBOE_INDEXES.rename(
        columns={
            "calc_end_time": "close_time",
            "calc_start_time": "open_time",
            "index_symbol": "symbol",
            "mkt_data_delay": "data_delay",
        },
    )

    indices_order: List[str] = [
        "symbol",
        "name",
        "description",
        "currency",
        "open_time",
        "close_time",
        "tick_days",
        "tick_frequency",
        "tick_period",
        "time_zone",
        "data_delay",
    ]

    CBOE_INDEXES = pd.DataFrame(CBOE_INDEXES, columns=indices_order).set_index("symbol")

    idx = (
        (CBOE_INDEXES.time_zone.str.contains("America/Chicago"))
        & (CBOE_INDEXES.currency.str.contains("USD"))
        & (CBOE_INDEXES.description.str.contains("Daily Risk Control") == 0)
        & (CBOE_INDEXES.name.str.contains("OEX%-SML% INDEX") == 0)
        & (CBOE_INDEXES.name.str.contains("Defined Risk Volatility Income") == 0)
        & (CBOE_INDEXES.description.str.contains("is designed to track") == 0)
    )

    return CBOE_INDEXES[idx]


def get_ticker_info(symbol: str, **kwargs) -> Dict[str, Any]:
    symbol = symbol.upper()
    SYMBOLS = get_cboe_directory()
    INDEXES = get_cboe_index_directory()
    data: Dict[str, Any] = {}

    if symbol not in SYMBOLS.index and symbol not in INDEXES.index:
        raise RuntimeError(f"Data not found for: {symbol}")

    url = (
        f"https://cdn.cboe.com/api/global/delayed_quotes/quotes/_{symbol}.json"
        if symbol in INDEXES.index or symbol in TICKER_EXCEPTIONS
        else f"https://cdn.cboe.com/api/global/delayed_quotes/quotes/{symbol}.json"
    )

    r = requests.get(url, timeout=10)

    if r.status_code not in set([200, 403]):
        raise requests.HTTPError(r.status_code)

    if r.status_code == 403:
        raise RuntimeError(f"Data was not found for, {symbol}.")

    data = r.json()["data"]

    if symbol in SYMBOLS.index.to_list():
        data.update({"name": SYMBOLS.at[symbol, "name"]})
    if symbol in INDEXES.index.to_list():
        data.update({"name": INDEXES.at[symbol, "name"]})

    _data = pd.DataFrame.from_dict(data, orient="index").transpose()

    _data = _data.rename(
        columns={
            "current_price": "price",
            "prev_day_close": "prev_close",
            "price_change": "change",
            "price_change_percent": "change_percent",
            "last_trade_time": "last_trade_timestamp",
            "security_type": "type",
        }
    )

    return _data.transpose()[0].to_dict()


def get_ticker_iv(symbol: str, **kwargs) -> Dict[str, float]:
    """Get annualized high/low historical and implied volatility over 30/60/90 day windows.

    Parameters
    ----------
    symbol: str
        The loaded ticker

    Returns
    -------
    pd.DataFrame: ticker_iv
    """

    symbol = symbol.upper()

    INDEXES = get_cboe_index_directory().index.to_list()

    quotes_iv_url = (
        "https://cdn.cboe.com/api/global/delayed_quotes/historical_data/_"
        f"{symbol}"
        ".json"
        if symbol in TICKER_EXCEPTIONS or symbol in INDEXES
        else f"https://cdn.cboe.com/api/global/delayed_quotes/historical_data/{symbol}.json"
    )

    h_iv = requests.get(quotes_iv_url, timeout=10)

    if h_iv.status_code not in set([200, 403]):
        raise requests.HTTPError(h_iv.status_code)

    if h_iv.status_code == 403:
        raise RuntimeError(f"Data was not found for, {symbol}.")

    data = pd.DataFrame(h_iv.json())[2:-1]["data"].rename(f"{symbol}")

    return data.to_dict()


def list_futures(**kwargs) -> List[dict]:
    """List of CBOE futures and their underlying symbols.

    Returns
    --------
    pd.DataFrame
        Pandas DataFrame with results.
    """

    r = requests.get(
        "https://cdn.cboe.com/api/global/delayed_quotes/symbol_book/futures-roots.json",
        timeout=10,
    )

    if r.status_code != 200:
        raise requests.HTTPError(r.status_code)

    return r.json()["data"]


def get_settlement_prices(
    settlement_date: Optional[date] = None,
    options: bool = False,
    archives: bool = False,
    final_settlement: bool = False,
    **kwargs,
) -> pd.DataFrame:
    """Gets the settlement prices of CBOE futures.

    Parameters
    -----------
    settlement_date: Optional[date]
        The settlement date. Only valid for active contracts. [YYYY-MM-DD]
    options: bool
        If true, returns options on futures.
    archives: bool
        Settlement price archives for select years and products.  Overridden by other parameters.
    final_settlement: bool
        Final settlement prices for expired contracts.  Overrides archives.

    Returns
    -------
    pd.DataFrame
        Pandas DataFrame with results.
    """
    url = ""
    if archives is True:
        url = "https://cdn.cboe.com/resources/futures/archive/volume-and-price/CFE_FinalSettlement_Archive.csv"

    if settlement_date is not None:
        url = f"https://www.cboe.com/us/futures/market_statistics/settlement/csv?dt={settlement_date}"
        if options is True:
            url = f"https://www.cboe.com/us/futures/market_statistics/settlement/csv?options=t&dt={settlement_date}"

    if settlement_date is None:
        url = "https://www.cboe.com/us/futures/market_statistics/settlement/csv"
        if options is True:
            url = "https://www.cboe.com/us/futures/market_statistics/settlement/csv?options=t"

    if final_settlement is True:
        url = "https://www.cboe.com/us/futures/market_statistics/final_settlement_prices/csv/"

    r = requests.get(url, timeout=10)

    if r.status_code != 200:
        raise RuntimeError(r.status_code)

    data = pd.read_csv(BytesIO(r.content), index_col=None, parse_dates=True)

    if data.empty:
        error_string = (
            f"No results found for, {settlement_date}."
            if settlement_date is not None
            else "No results found."
        )
        raise RuntimeError(error_string)

    data.columns = [to_snake_case(c) for c in data.columns]
    data = data.rename(columns={"expiration_date": "expiration"})

    if len(data) > 0:
        return data

    return pd.DataFrame()


class Europe:
    """Class for European CBOE data."""

    @staticmethod
    def get_all_index_definitions(**kwargs) -> Dict[Any, Any]:
        """Get the full list of European index definitions.

        Returns
        -------
        Dict[Any, Any]
            Dictionary with results.
        """

        r = cboe_session.get(
            "https://cdn.cboe.com/api/global/european_indices/definitions/all-definitions.json",
            timeout=10,
        )

        if r.status_code != 200:
            raise requests.HTTPError(r.status_code)
        return r.json()["data"]

    @staticmethod
    def list_indices(**kwargs) -> List[Dict]:
        """Gets names, currencies, ISINs, regions, and symbols for European indices.

        Returns
        -------
        Dict[str, str]
            List of dictionaries with the results.
        """

        data = Europe.get_all_index_definitions()
        data = (
            pd.DataFrame.from_records(pd.DataFrame(data)["index"])
            .drop(columns=["short_name"])
            .rename(columns={"long_name": "name"})
        )
        return data.to_dict("records")

    @staticmethod
    def list_index_constituents(symbol: str, **kwargs) -> List[str]:
        """List symbols for constituents of a European index.

        Parameters
        ----------
        symbol: str
            The symbol of the index.

        Returns
        -------
        List[str]
            List of constituents as ticker symbols.
        """

        SYMBOLS = pd.DataFrame(Europe.list_indices())["symbol"].to_list()
        symbol = symbol.upper()

        if symbol not in SYMBOLS:
            raise RuntimeError(
                f"The symbol, {symbol}, was not found in the CBOE European Index directory.",
            )

        url = f"https://cdn.cboe.com/api/global/european_indices/definitions/{symbol}.json"
        r = requests.get(url, timeout=10)

        if r.status_code != 200:
            raise requests.HTTPError(r.status_code)

        r_json = r.json()["constituents"]

        return [r_json[i]["constituent_symbol"] for i in range(0, len(r_json))]



"""European Indices End of Day fetcher for CBOE."""

from datetime import datetime
from typing import Any, Dict, List, Literal, Optional

import pandas as pd
from dateutil.relativedelta import relativedelta
from openbb_cboe.utils.helpers import Europe
from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.european_indices import (
    EuropeanIndicesData,
    EuropeanIndicesQueryParams,
)
from openbb_provider.utils.helpers import make_request
from pydantic import Field, field_validator


class CboeEuropeanIndicesQueryParams(EuropeanIndicesQueryParams):
    """CBOE European Indices End of Day query.

    Source: https://www.cboe.com/europe/indices/
    """

    interval: Optional[Literal["1d", "1m"]] = Field(
        description="Data granularity.",
        default="1d",
    )


class CboeEuropeanIndicesData(EuropeanIndicesData):
    """CBOE Stocks End of Day Data."""

    open: Optional[float] = Field(
        default=None,
        description="Opening price for the interval. Only valid when interval is 1m.",
    )
    high: Optional[float] = Field(
        default=None,
        description="High price for the interval. Only valid when interval is 1m.",
    )
    low: Optional[float] = Field(
        default=None,
        description="Low price for the interval. Only valid when interval is 1m.",
    )
    utc_datetime: Optional[datetime] = Field(
        default=None, description="UTC datetime. Only valid when interval is 1m."
    )

    @field_validator("date", mode="before", check_fields=False)
    def date_validate(cls, v):  # pylint: disable=E0213
        """Return datetime object from string."""
        try:
            return datetime.strptime(v, "%Y-%m-%d")
        except Exception:
            return datetime.strptime(v, "%Y-%m-%dT%H:%M:%S")


class CboeEuropeanIndicesFetcher(
    Fetcher[
        CboeEuropeanIndicesQueryParams,
        List[CboeEuropeanIndicesData],
    ]
):
    """Transform the query, extract and transform the data from the CBOE endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> CboeEuropeanIndicesQueryParams:
        """Transform the query."""
        now = datetime.now().date()
        transformed_params = params
        if params.get("start_date") is None:
            transformed_params["start_date"] = now - relativedelta(years=1)

        if params.get("end_date") is None:
            transformed_params["end_date"] = now

        return CboeEuropeanIndicesQueryParams(**transformed_params)

    @staticmethod
    def extract_data(
        query: CboeEuropeanIndicesQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the CBOE endpoint."""
        data = pd.DataFrame()
        query.symbol = query.symbol.upper()
        SYMBOLS = pd.DataFrame(Europe.list_indices())["symbol"].to_list()

        if query.symbol not in SYMBOLS:
            raise RuntimeError(
                f"The symbol, {query.symbol},"
                "was not found in the CBOE European Indices directory. "
                "Use `available_indices(europe=True)` to see the full list of indices."
            )
        if query.interval == "1d":
            url = f"https://cdn.cboe.com/api/global/european_indices/index_history/{query.symbol}.json"

            r = make_request(url)
            if r.status_code != 200:
                raise RuntimeError(r.status_code)

            r_json = r.json()

            data = pd.DataFrame.from_records(r_json["data"]).set_index("date")

            data.index = pd.to_datetime(data.index, format="%Y-%m-%d")

            if query.start_date is not None:
                query.end_date = (
                    query.end_date if query.end_date else datetime.now().date()
                )

                data = data[
                    (data.index >= pd.to_datetime(query.start_date, format="%Y-%m-%d"))
                    & (data.index <= pd.to_datetime(query.end_date, format="%Y-%m-%d"))
                ]
            data.index = data.index.astype(str)
            data = data.reset_index()

        if query.interval == "1m":
            url = f"https://cdn.cboe.com/api/global/european_indices/intraday_chart_data/{query.symbol}.json"

            r = make_request(url)
            if r.status_code != 200:
                raise RuntimeError(r.status_code)

            r_json = r.json()["data"]

            data = pd.DataFrame.from_records(pd.DataFrame(r_json)["price"])
            data["date"] = pd.DataFrame(r_json)["datetime"]
            data["utc_datetime"] = pd.to_datetime(pd.DataFrame(r_json)["utc_datetime"])

            data = data[["utc_datetime", "date", "open", "high", "low", "close"]].round(
                2
            )

        return data.to_dict("records")

    @staticmethod
    def transform_data(
        query: CboeEuropeanIndicesQueryParams, data: List[Dict], **kwargs: Any
    ) -> List[CboeEuropeanIndicesData]:
        """Transform the data to the standard format."""
        return [CboeEuropeanIndicesData.model_validate(d) for d in data]



"""Available Indices fetcher for CBOE"""

from datetime import time
from typing import Any, Dict, List, Optional

from openbb_cboe.utils.helpers import Europe, get_cboe_index_directory
from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.available_indices import (
    AvailableIndicesData,
    AvailableIndicesQueryParams,
)
from pydantic import Field


class CboeAvailableIndicesQueryParams(AvailableIndicesQueryParams):
    """CBOE Available Indices query.

    Source: https://www.cboe.com/europe/indices/
    """

    europe: bool = Field(
        description="Filter for European indices. False for US indices.", default=False
    )


class CboeAvailableIndicesData(AvailableIndicesData):
    """CBOE Available Indices data.

    Source: https://www.cboe.com/europe/indices/
    """

    isin: Optional[str] = Field(
        default=None,
        description="ISIN code for the index. Valid only for European indices.",
    )

    region: Optional[str] = Field(
        default=None,
        description="Region for the index. Valid only for European indices",
    )

    symbol: Optional[str] = Field(description="Symbol for the index.")

    description: Optional[str] = Field(
        default=None,
        description="Description for the index. Valid only for US indices.",
    )

    data_delay: Optional[int] = Field(
        default=None, description="Data delay for the index. Valid only for US indices."
    )

    open_time: Optional[time] = Field(
        default=None,
        description="Opening time for the index. Valid only for US indices.",
    )

    close_time: Optional[time] = Field(
        default=None,
        description="Closing time for the index. Valid only for US indices.",
    )

    time_zone: Optional[str] = Field(
        default=None, description="Time zone for the index. Valid only for US indices."
    )

    tick_days: Optional[str] = Field(
        default=None,
        description="The trading days for the index. Valid only for US indices.",
    )

    tick_frequency: Optional[str] = Field(
        default=None,
        description="The frequency of the index ticks. Valid only for US indices.",
    )

    tick_period: Optional[str] = Field(
        default=None,
        description="The period of the index ticks. Valid only for US indices.",
    )


class CboeAvailableIndicesFetcher(
    Fetcher[
        CboeAvailableIndicesQueryParams,
        List[CboeAvailableIndicesData],
    ]
):
    """Transform the query, extract and transform the data from the CBOE endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> CboeAvailableIndicesQueryParams:
        """Transform the query params."""
        return CboeAvailableIndicesQueryParams(**params)

    @staticmethod
    def extract_data(
        query: CboeAvailableIndicesQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the CBOE endpoint."""
        if query.europe is True:
            return Europe.list_indices()
        return get_cboe_index_directory().sort_index().reset_index().to_dict("records")

    @staticmethod
    def transform_data(
        query: CboeAvailableIndicesQueryParams, data: List[Dict], **kwargs: Any
    ) -> List[CboeAvailableIndicesData]:
        """Transform the data to the standard format."""
        return [CboeAvailableIndicesData.model_validate(d) for d in data]



"""CBOE Company Search fetcher."""

from typing import Any, Dict, List, Optional

from openbb_cboe.utils.helpers import get_cboe_directory
from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.equity_search import (
    EquitySearchData,
    EquitySearchQueryParams,
)
from pydantic import Field


class CboeEquitySearchQueryParams(EquitySearchQueryParams):
    """CBOE Company Search query.

    Source: https://www.cboe.com/
    """


class CboeEquitySearchData(EquitySearchData):
    """CBOE Company Search Data."""

    __alias_dict__ = {"name": "Company Name"}

    dpm_name: Optional[str] = Field(
        default=None,
        description="Name of the primary market maker.",
        alias="DPM Name",
    )
    post_station: Optional[str] = Field(
        default=None, description="Post and station location on the CBOE trading floor."
    )


class CboeEquitySearchFetcher(
    Fetcher[
        CboeEquitySearchQueryParams,
        List[CboeEquitySearchData],
    ]
):
    """Transform the query, extract and transform the data from the CBOE endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> CboeEquitySearchQueryParams:
        """Transform the query."""
        return CboeEquitySearchQueryParams(**params)

    @staticmethod
    def extract_data(  # pylint: disable=unused-argument
        query: CboeEquitySearchQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> Dict:
        """Return the raw data from the CBOE endpoint."""
        data = {}
        symbols = get_cboe_directory().reset_index().replace("nan", None)
        target = "name" if query.is_symbol is False else "symbol"
        idx = symbols[target].str.contains(query.query, case=False)
        result = symbols[idx].to_dict("records")
        data.update({"results": result})

        return data

    @staticmethod
    def transform_data(
        query: CboeEquitySearchQueryParams, data: Dict, **kwargs: Any
    ) -> List[CboeEquitySearchData]:
        """Transform the data to the standard format."""
        return [CboeEquitySearchData.model_validate(d) for d in data["results"]]



"""CBOE Equity Historical fetcher."""


from datetime import datetime, timedelta
from typing import Any, Dict, List, Literal, Optional

import pandas as pd
from openbb_cboe.utils.helpers import (
    TICKER_EXCEPTIONS,
    get_cboe_directory,
    get_cboe_index_directory,
    get_ticker_info,
)
from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.equity_historical import (
    EquityHistoricalData,
    EquityHistoricalQueryParams,
)
from openbb_provider.utils.descriptions import QUERY_DESCRIPTIONS
from openbb_provider.utils.helpers import make_request
from pydantic import Field


class CboeEquityHistoricalQueryParams(EquityHistoricalQueryParams):
    """CBOE Stock end of day query.

    Source: https://www.cboe.com/
    """

    interval: Optional[Literal["1m", "1d"]] = Field(
        default="1d",
        description=QUERY_DESCRIPTIONS.get("interval", ""),
    )


class CboeEquityHistoricalData(EquityHistoricalData):
    """CBOE Stocks End of Day Data."""

    calls_volume: Optional[float] = Field(
        default=None,
        description="Number of calls traded during the most recent trading period. Only valid if interval is 1m.",
    )
    puts_volume: Optional[float] = Field(
        default=None,
        description="Number of puts traded during the most recent trading period. Only valid if interval is 1m.",
    )
    total_options_volume: Optional[float] = Field(
        default=None,
        description="Total number of options traded during the most recent trading period. Only valid if interval is 1m.",
    )


class CboeEquityHistoricalFetcher(
    Fetcher[
        CboeEquityHistoricalQueryParams,
        List[CboeEquityHistoricalData],
    ]
):
    """Transform the query, extract and transform the data from the CBOE endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> CboeEquityHistoricalQueryParams:
        """Transform the query. Setting the start and end dates for a 1 year period."""
        return CboeEquityHistoricalQueryParams(**params)

    @staticmethod
    def extract_data(
        query: CboeEquityHistoricalQueryParams,  # pylint: disable=unused-argument
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the CBOE endpoint."""
        # Symbol directories are cached for seven days and are used for error handling and URL generation.
        SYMBOLS = get_cboe_directory()
        INDEXES = get_cboe_index_directory().index.to_list()
        multi = pd.DataFrame()
        now = datetime.now()
        tickers = (
            query.symbol.split(",") if "," in query.symbol else [query.symbol.upper()]
        )
        if len(tickers) > 1:
            query.start_date = (
                query.start_date if query.start_date else now - timedelta(days=720)
            )
        if len(tickers) == 1:
            query.start_date = (
                query.start_date if query.start_date else now - timedelta(days=50000)
            )
        query.end_date = query.end_date if query.end_date else now

        def get_one(symbol, start_date, end_date, interval):
            symbol = symbol.upper()
            if "^" in symbol:
                symbol = symbol.replace("^", "")
            if len(tickers) == 1 and symbol == "NDX":
                raise RuntimeError("NDX is not supported by this function.")

            data = pd.DataFrame()

            if symbol not in SYMBOLS.index and symbol not in INDEXES:
                raise RuntimeError(
                    f"The symbol, {symbol}, was not found in the CBOE directory.  Use `search()`."
                )

            def __generate_historical_prices_url(
                symbol,
                data_type: Optional[Literal["intraday", "historical"]] = "historical",
            ) -> str:
                """Generate the final URL for historical prices data."""
                url: str = (
                    f"https://cdn.cboe.com/api/global/delayed_quotes/charts/{data_type}"
                )
                url = (
                    url + f"/_{symbol}.json"
                    if symbol in TICKER_EXCEPTIONS or symbol in INDEXES
                    else url + f"/{symbol}.json"
                )
                return url

            url = (
                __generate_historical_prices_url(symbol, "intraday")
                if interval == "1m"
                else __generate_historical_prices_url(symbol)
            )
            r = make_request(url)

            if r.status_code != 200:
                raise RuntimeError(r.status_code)

            if interval == "1d":
                data = (
                    pd.DataFrame(r.json()["data"])[
                        ["date", "open", "high", "low", "close", "volume"]
                    ]
                ).set_index("date")

                # Fill in missing data from current or most recent trading session.

                today = pd.to_datetime(datetime.now().date())
                if today.weekday() > 4:
                    day_minus = today.weekday() - 4
                    today = pd.to_datetime(today - timedelta(days=day_minus))
                if today != data.index[-1]:
                    _today = pd.Series(get_ticker_info(symbol))
                    today_df = pd.Series(dtype="object")
                    today_df["open"] = round(_today["open"], 2)
                    today_df["high"] = round(_today["high"], 2)
                    today_df["low"] = round(_today["low"], 2)
                    today_df["close"] = round(_today["close"], 2)
                    if symbol not in INDEXES and symbol not in TICKER_EXCEPTIONS:
                        data = data[data["volume"] > 0]
                        today_df["volume"] = _today["volume"]
                    today_df["date"] = today.date()
                    today_df = pd.DataFrame(today_df).transpose().set_index("date")

                    data = pd.concat([data, today_df], axis=0)

                # If ticker is an index there is no volume data and the types must be set.

                if symbol in INDEXES or symbol in TICKER_EXCEPTIONS:
                    data = data[["open", "high", "low", "close", "volume"]]
                    data["open"] = round(data.open.astype(float), 2)
                    data["high"] = round(data.high.astype(float), 2)
                    data["low"] = round(data.low.astype(float), 2)
                    data["close"] = round(data.close.astype(float), 2)
                    data["volume"] = 0

                data.index = pd.to_datetime(data.index)
                data = data[data["open"] > 0]
                data = data[
                    (data.index >= pd.to_datetime(start_date))
                    & (data.index <= pd.to_datetime(end_date))
                ]
            if query.interval == "1m":
                data_list = r.json()["data"]
                date: list[datetime] = []
                open_: list[float] = []
                high: list[float] = []
                low: list[float] = []
                close: list[float] = []
                volume: list[float] = []
                calls_volume: list[float] = []
                puts_volume: list[float] = []
                total_options_volume: list[float] = []

                for data in data_list:
                    date.append(data["datetime"])
                    open_.append(data["price"]["open"])
                    high.append(data["price"]["high"])
                    low.append(data["price"]["low"])
                    close.append(data["price"]["close"])
                    volume.append(data["volume"]["stock_volume"])
                    calls_volume.append(data["volume"]["calls_volume"])
                    puts_volume.append(data["volume"]["puts_volume"])
                    total_options_volume.append(data["volume"]["total_options_volume"])

                data = pd.DataFrame()
                date = [d.replace("T", " ") for d in date]
                date = [datetime.strptime(d, "%Y-%m-%d %H:%M:%S") for d in date]
                data["date"] = date
                data["open"] = open_
                data["high"] = high
                data["low"] = low
                data["close"] = close
                data["volume"] = volume
                data["calls_volume"] = calls_volume
                data["puts_volume"] = puts_volume
                data["total_options_volume"] = total_options_volume
                data = data.set_index("date").sort_index()
                data.index = data.index.strftime("%Y-%m-%d %H:%M:%S")
                data = data[data["open"] > 0]

            data["volume"] = round(data["volume"].astype("int64"))
            data["symbol"] = symbol
            return data.drop_duplicates().reset_index()

        for ticker in tickers:
            if ticker != "NDX":
                data = get_one(
                    ticker,
                    start_date=query.start_date,
                    end_date=query.end_date,
                    interval=query.interval,
                ).set_index(["date", "symbol"])
                multi = pd.concat([multi, data], axis=0)

        multi = multi.sort_index().reset_index()
        if len(tickers) == 1:
            multi = multi.drop(columns=["symbol"])

        return multi.to_dict("records")

    @staticmethod
    def transform_data(
        query: CboeEquityHistoricalQueryParams, data: List[Dict], **kwargs: Any
    ) -> List[CboeEquityHistoricalData]:
        """Transform the data to the standard format."""
        return [CboeEquityHistoricalData.model_validate(d) for d in data]



"""CBOE provider module."""


from openbb_cboe.models.available_indices import CboeAvailableIndicesFetcher
from openbb_cboe.models.equity_historical import CboeEquityHistoricalFetcher
from openbb_cboe.models.equity_info import CboeEquityInfoFetcher
from openbb_cboe.models.equity_search import CboeEquitySearchFetcher
from openbb_cboe.models.european_index_constituents import (
    CboeEuropeanIndexConstituentsFetcher,
)
from openbb_cboe.models.european_indices import (
    CboeEuropeanIndicesFetcher,
)
from openbb_cboe.models.futures_curve import CboeFuturesCurveFetcher
from openbb_cboe.models.index_search import CboeIndexSearchFetcher
from openbb_cboe.models.index_snapshots import CboeIndexSnapshotsFetcher
from openbb_cboe.models.market_indices import (
    CboeMarketIndicesFetcher,
)
from openbb_cboe.models.options_chains import CboeOptionsChainsFetcher
from openbb_provider.abstract.provider import Provider

cboe_provider = Provider(
    name="cboe",
    website="https://www.cboe.com/",
    description="""Cboe is the world's go-to derivatives and exchange network,
    delivering cutting-edge trading, clearing and investment solutions to people
    around the world.""",
    required_credentials=None,
    fetcher_dict={
        "EquitySearch": CboeEquitySearchFetcher,
        "OptionsChains": CboeOptionsChainsFetcher,
        "EquityHistorical": CboeEquityHistoricalFetcher,
        "EquityInfo": CboeEquityInfoFetcher,
        "FuturesCurve": CboeFuturesCurveFetcher,
        "AvailableIndices": CboeAvailableIndicesFetcher,
        "EuropeanIndexConstituents": CboeEuropeanIndexConstituentsFetcher,
        "EuropeanIndices": CboeEuropeanIndicesFetcher,
        "MarketIndices": CboeMarketIndicesFetcher,
        "IndexSearch": CboeIndexSearchFetcher,
        "IndexSnapshots": CboeIndexSnapshotsFetcher,
    },
)



"""CBOE Index Search fetcher."""

from datetime import time
from typing import Any, Dict, List, Optional

import pandas as pd
from openbb_cboe.utils.helpers import Europe, get_cboe_index_directory
from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.index_search import (
    IndexSearchData,
    IndexSearchQueryParams,
)
from pydantic import Field


class CboeIndexSearchQueryParams(IndexSearchQueryParams):
    """CBOE Index Search query.  Search the CBOE company directory by name or ticker.

    Source: https://www.cboe.com/
    """

    europe: bool = Field(
        description="Filter for European indices. False for US indices.", default=False
    )


class CboeIndexSearchData(IndexSearchData):
    """CBOE Company Search Data."""

    isin: Optional[str] = Field(
        description="ISIN code for the index. Valid only for European indices.",
        default=None,
    )
    region: Optional[str] = Field(
        description="Region for the index. Valid only for European indices",
        default=None,
    )
    description: Optional[str] = Field(
        description="Description for the index.", default=None
    )
    data_delay: Optional[int] = Field(
        description="Data delay for the index. Valid only for US indices.", default=None
    )
    currency: Optional[str] = Field(description="Currency for the index.", default=None)
    time_zone: Optional[str] = Field(
        description="Time zone for the index. Valid only for US indices.", default=None
    )
    open_time: Optional[time] = Field(
        description="Opening time for the index. Valid only for US indices.",
        default=None,
    )
    close_time: Optional[time] = Field(
        description="Closing time for the index. Valid only for US indices.",
        default=None,
    )
    tick_days: Optional[str] = Field(
        description="The trading days for the index. Valid only for US indices.",
        default=None,
    )
    tick_frequency: Optional[str] = Field(
        description="Tick frequency for the index. Valid only for US indices.",
        default=None,
    )
    tick_period: Optional[str] = Field(
        description="Tick period for the index. Valid only for US indices.",
        default=None,
    )


class CboeIndexSearchFetcher(
    Fetcher[
        CboeIndexSearchQueryParams,
        List[CboeIndexSearchData],
    ]
):
    """Transform the query, extract and transform the data from the CBOE endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> CboeIndexSearchQueryParams:
        """Transform the query."""
        return CboeIndexSearchQueryParams(**params)

    @staticmethod
    def extract_data(
        query: CboeIndexSearchQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the CBOE endpoint."""
        symbols = pd.DataFrame()
        if query.europe is True:
            symbols = pd.DataFrame(Europe.list_indices())
        if query.europe is False:
            symbols = get_cboe_index_directory().reset_index()

        target = "name" if not query.is_symbol else "symbol"
        idx = symbols[target].str.contains(query.query, case=False)
        result = symbols[idx]

        return result.to_dict("records")

    @staticmethod
    def transform_data(
        query: CboeIndexSearchQueryParams, data: dict, **kwargs: Any
    ) -> List[CboeIndexSearchData]:
        """Transform the data to the standard format."""
        return [CboeIndexSearchData.model_validate(d) for d in data]



"""CBOE Equity Info fetcher."""

import concurrent.futures
from datetime import datetime
from typing import Any, Dict, List, Optional

import pandas as pd
from openbb_cboe.utils.helpers import (
    get_cboe_directory,
    get_cboe_index_directory,
    get_ticker_info,
    get_ticker_iv,
)
from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.equity_info import (
    EquityInfoData,
    EquityInfoQueryParams,
)
from pydantic import Field


class CboeEquityInfoQueryParams(EquityInfoQueryParams):
    """CBOE Company Search query.

    Source: https://www.cboe.com/
    """


class CboeEquityInfoData(EquityInfoData):
    """CBOE Company Search Data."""

    type: Optional[str] = Field(default=None, description="Type of asset.")
    exchange_id: Optional[int] = Field(
        default=None, description="The Exchange ID number."
    )
    tick: Optional[str] = Field(
        default=None, description="Whether the last sale was an up or down tick."
    )
    bid: Optional[float] = Field(default=None, description="Current bid price.")
    bid_size: Optional[float] = Field(default=None, description="Bid lot size.")
    ask: Optional[float] = Field(default=None, description="Current ask price.")
    ask_size: Optional[float] = Field(default=None, description="Ask lot size.")
    volume: Optional[float] = Field(
        default=None, description="Stock volume for the current trading day."
    )
    iv30: Optional[float] = Field(
        default=None, description="The 30-day implied volatility of the stock."
    )
    iv30_change: Optional[float] = Field(
        default=None, description="Change in 30-day implied volatility of the stock."
    )
    last_trade_timestamp: Optional[datetime] = Field(
        default=None, description="Last trade timestamp for the stock."
    )
    iv30_annual_high: Optional[float] = Field(
        default=None, description="The 1-year high of implied volatility."
    )
    hv30_annual_high: Optional[float] = Field(
        default=None, description="The 1-year high of realized volatility."
    )
    iv30_annual_low: Optional[float] = Field(
        default=None, description="The 1-year low of implied volatility."
    )
    hv30_annual_low: Optional[float] = Field(
        default=None, description="The 1-year low of realized volatility."
    )
    iv60_annual_high: Optional[float] = Field(
        default=None, description="The 60-day high of implied volatility."
    )
    hv60_annual_high: Optional[float] = Field(
        default=None, description="The 60-day high of realized volatility."
    )
    iv60_annual_low: Optional[float] = Field(
        default=None, description="The 60-day low of implied volatility."
    )
    hv60_annual_low: Optional[float] = Field(
        default=None, description="The 60-day low of realized volatility."
    )
    iv90_annual_high: Optional[float] = Field(
        default=None, description="The 90-day high of implied volatility."
    )
    hv90_annual_high: Optional[float] = Field(
        default=None, description="The 90-day high of realized volatility."
    )


class CboeEquityInfoFetcher(
    Fetcher[
        CboeEquityInfoQueryParams,
        List[CboeEquityInfoData],
    ]
):
    """Transform the query, extract and transform the data from the CBOE endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> CboeEquityInfoQueryParams:
        """Transform the query."""
        return CboeEquityInfoQueryParams(**params)

    @staticmethod
    def extract_data(
        query: CboeEquityInfoQueryParams,  # pylint: disable=unused-argument
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the CBOE endpoint."""
        results = []

        INDEXES = get_cboe_index_directory().index.to_list()
        SYMBOLS = get_cboe_directory()

        def get_one(symbol):
            data = pd.Series(dtype="object")
            if symbol in SYMBOLS.index or symbol in INDEXES:
                _info = pd.Series(get_ticker_info(symbol))
                _iv = pd.Series(get_ticker_iv(symbol))
                data = (
                    pd.DataFrame(pd.concat([_info, _iv]))
                    .transpose()
                    .drop(columns="seqno")
                    .iloc[0]
                )

                results.append(data.to_dict())

        with concurrent.futures.ThreadPoolExecutor() as executor:
            executor.map(get_one, query.symbol.split(","))

        return (
            pd.DataFrame.from_records(results)
            .sort_values(by="symbol")
            .to_dict("records")
        )

    @staticmethod
    def transform_data(
        query: CboeEquityInfoQueryParams, data: List[Dict], **kwargs: Any
    ) -> List[CboeEquityInfoData]:
        """Transform the data to the standard format."""
        return [CboeEquityInfoData.model_validate(d) for d in data]
